// ----------------------------------------------------------------------------
// CVCL.java
// CVC Lite Interface
// $Id: CVCL.java,v 1.31 2009/10/29 12:07:25 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving.CVCL;

import java.io.*;
import java.util.*;
import fmrisc.Proving.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;
import fmrisc.ProofNavigator.*;

/**
 * 
 * CVC Lite interface
 *
 */
public final class CVCL implements Prover
{
  /****************************************************************************
   * 
   * Static methods
   * 
   ***************************************************************************/
  
  /****************************************************************************
   * create CVCL identifier from given name
   * @param ident an identifier
   * @return an identifier usable for CVCL
   **************************************************************************/
  public static String getExternalName(ValueDeclIdentifier ident)
  {
    // CVCL has problems with identical identifiers in nested quantifiers
    if (ident.getSymbol().getDepth() == 0)
      // global names take their externally visible name
      return ident.getName() + "__";
    else
      // local names take their unique name
      return ident.getUniqueName() + "__";
  }
  
  /***************************************************************************
   * Get original name from external CVCL name.
   * @param name a CVCL name (may be returned by getExternalName())
   * @return the original name (may be name itself)
   **************************************************************************/
  public static String getOriginalName(String name)
  {
    int pos = name.indexOf("__");
    if (pos == -1) return name;
    return name.substring(0, pos);
  }
  
  private String command;     // the invokation command
  private PrintWriter log;    // logfile (may be null) 
  private Process process;    // the CVCL instance
  private PrintWriter out;    // its output stream
  private Reader in;          // its input stream
  private Reader err;         // its error stream
  private Reader reader;      // the current reader stream
  private String error;       // error message 
  private WatchDog watchDog;  // thread guarding the CVCL instance
  private Vector context = new Vector(); // the context set up so far for CVCL
  private boolean generate;   // counter examples to be generated
  private boolean useMask;      // true if mask is to be used
  private CVCLFormulaMask mask; // mask used for formulas
  
  /***************************************************************************
   * Create CVCL instance by call of command
   * @param command command line string to invoke CVCL
   * @param log stream to log commands to (null, if none)
   * @param useMask true iff quantified formulas are to be masked from CVCL.
   **************************************************************************/
  public CVCL(String command, PrintWriter log, boolean useMask)
  {
    this.command = command;
    this.log = log;
    this.generate = false;
    this.useMask = useMask;
    this.mask = new CVCLFormulaMask();
    boolean okay = connect();
    if (!okay)
    {
      // wait a bit and try a second time
      try { Thread.sleep(3000); } catch(InterruptedException e) { }
      okay = connect();
      if (!okay)
      {
        setError("process could not be started.");
        return;
      }
    }
    watchDog = new WatchDog();
  }
  
  /***************************************************************************
   * 
   * State access methods
   * 
   **************************************************************************/
  
  /***************************************************************************
   * true iff CVCL instance is okay
   * @return true iff CVCL instance is okay
   **************************************************************************/
  public boolean isOkay()
  {
    return error == null;
  }
  
  /****************************************************************************
   * error message, if CVCL instance is not okay
   * @return error message, if !isOkay()
   **************************************************************************/
  public String getError()
  {
    return error;
  }
  
  /***************************************************************************
   * set error message
   * @param message the error message
   **************************************************************************/
  public void setError(String message)
  {
    writeLog("error: " + message);
    error = message;
  }
  
  /***************************************************************************
   * generate counterexamples, if possible.
   * @param generate true iff counterexamples are to be generated
   **************************************************************************/
  public void generateCounterExamples(boolean generate)
  {
    this.generate = generate;
  }
  
  /***************************************************************************
   * Try hard to simplify (potentially timeconsuming).
   * @param hard true iff prover is to try hard.
   **************************************************************************/
  public void tryHard(boolean hard)
  {
    this.useMask = !hard;
  }
  
  /***************************************************************************
   *
   * Synchronized methods (to be invoked by multiple threads).
   * 
   **************************************************************************/
  
  /***************************************************************************
   * write message to log file
   * @param message to be written
   **************************************************************************/
  private synchronized void writeLog(String message)
  {
    if (log == null) return;
    log.println(new Date() + ": " + message);
    log.flush();
  }
  
  /***************************************************************************
   * Watchdog thread to guard process instance.
   * Thread is automatically started on creation.
   * If watchdog is not alive, process has died.
   **************************************************************************/
  private class WatchDog extends Thread
  {
    private boolean terminated = false;
    
    public WatchDog()
    {
      start();
    }
    
    public boolean isTerminated()
    {
      return terminated;
    }
    
    public void run()
    {
      try
      {
        while (true)
        {
          int code = process.waitFor();
          terminated = watchDog == null;
          if (terminated) return;
          writeLog("process terminated with result code " + code);
          boolean okay = restart();
          if (!okay) return;
        }
      }
      catch(InterruptedException e)
      {
        setError("InterruptedException: " + e.toString());
      }
    }
  }
  
  /****************************************************************************
   * create and connect to a CVCL instance
   * @return true iff connection was successful
   **************************************************************************/
  private synchronized boolean connect()
  {
    try
    {
      writeLog("executing " + command);
      process = Runtime.getRuntime().exec(command);
      if (process == null) return false;
      out = new PrintWriter(process.getOutputStream(), true);
      in = new InputStreamReader(process.getInputStream());
      reader = in;
      err = new InputStreamReader(process.getErrorStream());
      String line = readLine();
      if (line == null)
      {
        setError("Process does not create output.");
        return false;
      }
      if (!line.equals("CVC> "))
      {
        setError("Process does not write CVC> but \"" + line + "\"");
        return false;
      }
      writeLog("process started");
      return true;
    }
    catch (IOException e)
    {
      setError("IOException(connect): " + e.toString());
      return false;
    }
  }
  
  /***************************************************************************
   * Restart prover process and send context.
   * @return true if restart was successful.
   **************************************************************************/
  private synchronized boolean restart()
  {
    writeLog("restarting process");
    boolean okay = connect();
    if (!okay) return false;
    int n = context.size();
    for (int i=0; i<n; i++)
    {
      send((String)context.elementAt(i));
      flush();
      ignoreAnswer();
    }
    try
    {
      int code = process.exitValue();
      writeLog("process terminated with result code " + code);
      writeLog("giving up to restart process");
      return false;
    }
    catch(IllegalThreadStateException e)
    { }
    error = null;
    writeLog("process restarted");
    return true;
  }
  
  /***************************************************************************
   * Terminate prover instance.
   **************************************************************************/
  public synchronized void terminate()
  {
    WatchDog w = watchDog;
    
    // signal watchdog not to restart process but to terminate
    watchDog = null; 
    process.destroy();
    setError("process is terminated by user");
   
    // wait for watchdog to terminate (signalling termination of prover)
    if (w != null) try { w.join(); } catch(InterruptedException e) { System.out.println("interrupted"); }
  }

  
  /***************************************************************************
   * Abort current prover activity. 
   * @return true iff new prover could be successfully started.
   **************************************************************************/
  public synchronized boolean abort()
  { 
    writeLog("aborting process");
    State.setAborted(true);
    terminate();
    int attempts = 2;
    while (true)
    {    
      try { Thread.sleep(1000); } catch(InterruptedException e) { }
      boolean okay = restart();
      if (okay) break;
      attempts = attempts-1;
      if (attempts == 0) return false;
      writeLog("Could not restart process, try again ...");
    }
    watchDog = new WatchDog();
    return true;
  }
  
  /****************************************************************************
   * Send message to CVCL instance.
   * @param message the message to be sent to the CVCL instance.
   ***************************************************************************/
  private synchronized void send(String message)
  {
    writeLog("send: " + message);
    out.print(message);
  }

  /****************************************************************************
   * Flush messages to CVCL instance; must be called before any response
   * can be expected. We need this function since some actions need multiple
   * sends before a response should occur.
   ***************************************************************************/
  private synchronized void flush()
  {
    // autoflushing enabled for stream
    out.println();
  }
  
  /***************************************************************************
   * add command to context
   * @param command the command to be added
   *************************************************************************/
  private synchronized void addContext(String command)
  {
    context.add(command);
    send(command);
    flush();
  }
  
 /***************************************************************************
  *
  * Auxiliary methods (mostly private).
  * 
  **************************************************************************/
  
  /****************************************************************************
   * return CVCL representation of declaration
   * @param decl (value or type) declaration
   * @return the CVCL representation of decl
   ***************************************************************************/
  private String toCVCL(Declaration decl)
  {
    // ASTUtil.setPrintCVCL(true);
    // String result = decl.toString();
    // ASTUtil.setPrintCVCL(false);
    // return result;
    return CVCLPrinter.toString(decl);
  }
  
  /****************************************************************************
   * return CVCL representation of expression
   * @param exp an expression (formula or term)
   * @return the CVCL representation of exp
   ***************************************************************************/
  public String toCVCL(Expression exp)
  {
    if (useMask) 
    {
      mask.resetSymbols();
      State.setFormulaMask(mask);
    }
    // ASTUtil.setPrintCVCL(true);
    // String result = exp.toString();
    // ASTUtil.setPrintCVCL(false);
    String result = CVCLPrinter.toString(Rewriting.rewrite(exp));
    if (useMask)
    {
      State.setFormulaMask(null);
      Iterator iterator = mask.getSymbols().iterator();
      while (iterator.hasNext())
      {
        ValueSymbol symbol = (ValueSymbol)(iterator.next());
        addDeclaration(Construct.valueDeclaration(
          symbol.getValueDeclIdentifier(), 
          symbol.getType(),
          null));
      }
    }
    return result;
  }
  
  /****************************************************************************
   * read as much input as possible from process
   * but until end of line at most 
   * @return the text read from the process
   **************************************************************************/
  private String readLine()
  {
    try
    {
      StringBuffer buffer = new StringBuffer();
      int ch = reader.read();  // block until at least one character available
      if (ch == -1 && reader == in)
      {
        reader = err;
        ch = reader.read();
      }
      if (ch == -1) 
      {
        setError("End of input encountered (probably process aborted, see logfile).");
        return null;
      }
      do
      {
        if (ch == '\n') break;   // new line terminates input
        buffer.append((char)ch);
        if (!reader.ready()) break;
        ch = reader.read();
      }
      while (ch != -1);
      String string = buffer.toString();
      writeLog("received: " + string);
      return string;
    }
    catch(IOException e)
    {      
      setError("IOException(readline): " + e.toString());
      return null;
    }
  }
  
  /****************************************************************************
   * get message from CVCL instance
   * @return message the message received from the CVCL instance
   *         (null, if an error occured)
   ***************************************************************************/
  private String receive()
  {
    StringBuffer result = new StringBuffer();
    boolean empty = true;
    while (true)
    {
      // if (!watchDog.isAlive()) return watchDog.getStdError();
      String line = readLine();
      if (line == null) 
      {
        if (empty)
          return null;
        else
          return result.toString();
      }
      if (line.equals("CVC> ")) return result.toString();
      if (!empty) result.append('\n');
      empty = false;
      result.append(line);
    }
  }
  
  /***************************************************************************
   * Receive counterexample from CVCL interface.
   * This may fail depending on the theory used (which lets process terminate).
   **************************************************************************/
  private String receiveCounterExample()
  {
    // disable watchDog
    // WatchDog w = watchDog;
    // watchDog = null;
    
    // ask for counterexample
    send("COUNTEREXAMPLE;");
    flush();
    String counterExample = receive();
    
    // give watchdog a chance to terminate (race condition!)
    // try { Thread.sleep(1); } catch(InterruptedException e) { }
    
    // counterexample generation failed, restart process and watchdog
    // if (w.isTerminated())
    // {
    //  boolean okay = restart();
    //  if (okay) watchDog = new WatchDog();
    //  return null;
    // }
    
    // counterexample generation successful, enable watchdog again
    // watchDog = w;
    
    return counterExample;
  }
  
  /***************************************************************************
   * receive answer from CVCL interface
   **************************************************************************/
  private Answer receiveAnswer()
  {
    String message = receive();
    if (message == null) 
      return null;
    if (message.equals("Valid."))
      return new ValidAnswer("CVCL");
    if (generate && message.equals("InValid."))
      return new InvalidAnswerCVCL(this, receiveCounterExample());
    return new UnknownAnswer(message);
  }
  
  /***************************************************************************
   * ignore answer from CVCL interface
   **************************************************************************/
  private void ignoreAnswer()
  {
    String message = receive();
  }
  
  /***************************************************************************
   * Parse CVCL String as an expression.
   * @param string the CVCL representation of an expression.
   * @return the corresponding expression (null, if an error occured)
   **************************************************************************/
  public Expression parseExpression(String string)
  {
    try
    {
      CVCLLexer lexer = new CVCLLexer(this, new StringReader(string));
      CVCLParser parser = new CVCLParser(this, lexer);
      if (useMask) State.setFormulaMask(mask);
      AST result = parser.main();
      if (useMask) State.setFormulaMask(null);
      if (!(result instanceof Expression)) 
      {
        System.out.println("CVCL returned \"" + string +
          "\" which could not be parsed as an expression");
        State.setError("CVCL returned \"" + string +
          "\" which could not be parsed as an expression");
        return null;
      }
      return (Expression)result;
    }
    catch (antlr.ANTLRException e) 
    {
      setError("ANTLR exception: " + e);
      return null;
    }
  }
  
  /***************************************************************************
   * receive expression from CVLC instance
   * @return the expression (null if an error occured)
   **************************************************************************/
  private Expression receiveExpression()
  {
    String message = receive();
    if (message == null) return null;
    return parseExpression(message);
  }
  
 /***************************************************************************
  *
  * Public unsychronized access methods.
  * 
  **************************************************************************/
  
  /***************************************************************************
   * Add declaration.
   * @param decl declaration that is added to prover state.
   *************************************************************************/
  public void addDeclaration(Declaration decl)
  {
    addContext(toCVCL(decl) + ";");
    ignoreAnswer();
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration vdecl = (ValueDeclaration)decl;
      TypeExpression te = 
        vdecl.getValueDeclIdentifier().getSymbol().getClearType();
      if (te != null)
      {
        Expression formula = te.getExpression();
        if (formula != null) assertAxiom(formula);
      }
    }
  }
  
  /***************************************************************************
   * Start a new subcontext for formula assertions.
   *************************************************************************/
  public void pushContext()
  {
    addContext("PUSH;");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * End subcontext for formula assertions. All assertions since the last
   * pushContext() are undone.
   *************************************************************************/
  public void popContext()
  {
    addContext("POP;");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * assert axiom about previously asserted declarations.
   * @param axiom formula whose truth is asserted;
   *        declaration symbols must be only used in context of axiom.
   *************************************************************************/
  public void assertAxiom(Expression axiom)
  {
    // expose CVCL to unmasked formula
    // boolean use = useMask;
    // useMask = false;
    assertFormula(axiom);
    // useMask = use;
  }
  
  /***************************************************************************
   * Assert truth of formula.
   * @param formula formula whose truth is asserted.
   *************************************************************************/
  public void assertFormula(Expression formula)
  {
    addContext("ASSERT " + toCVCL(formula) + ";");
    ignoreAnswer();
  }
  
  /***************************************************************************
   * Query truth of formula.
   * @param formula formula whose truth is queried.
   * @return the prover answer (null, if a problem occured).
   *************************************************************************/
  public Answer queryFormula(Expression formula)
  {
    send("PUSH;");
    send("QUERY " + toCVCL(formula) + ";");
    flush();
    Answer answer=receiveAnswer();
    send("POP;");
    flush();
    ignoreAnswer();
    return answer;
  }

  /***************************************************************************
   * Simplify expression.
   * @param exp expression to be simplified.
   * @return the simplified expression (null, if a problem occured).
   *************************************************************************/
  public Expression simplifyExpression(Expression exp)
  {
    String exp0 = toCVCL(exp);
    send("TRANSFORM " + exp0 + ";");
    flush();
    Expression result = receiveExpression();
    if (result == null) return null;
    Type type = Checking.process(result);
    if (type == null) 
    {
      // do not give warnings because type checking may simply fail because 
      // exp0 contains weaker type declarations than exp
      // State.getSystemOut().println("Warning: CVCL returns mistyped simplification result.");
      // State.getSystemOut().println("Type checking message: " + State.getError());
      // State.resetError();
      // State.getSystemOut().println("CVCL result is ignored.");
      State.resetError();
      result = exp;
    }
    return Rewriting.rewrite(result);
  }
}
