// ----------------------------------------------------------------------------
// CVCLPrinter.java
// Printing abstract syntax trees in CVCL syntax.
// $Id: CVCLPrinter.java,v 1.12 2010/07/07 15:21:14 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving.CVCL;

import fmrisc.ProofNavigator.*;
import fmrisc.Syntax.*;
import fmrisc.Syntax.Number;
import fmrisc.Semantics.*;

import java.io.*;

/**
 * 
 * Printing abstract syntax trees in CVLC syntax.
 *
 */
public final class CVCLPrinter
{
  /***************************************************************************
   * Convert abstract syntax tree to CVCL syntax.
   * @param tree the tree to be converted.
   * @return the conversion result.
   **************************************************************************/
  public static String toString(AST tree)
  {
    StringWriter out = new StringWriter();
    print(tree, new PrintWriter(out));
    return out.toString();
  }
  
  /***************************************************************************
   * Print abstract syntax tree in CVCL syntax.
   * @param tree the tree to be printed.
   * @param out the stream to which the tree is printed.
   **************************************************************************/
  public static void print(AST tree, PrintWriter out)
  {
    new CVCLPrinter(out).print(tree);
  }
  
  // the stream to which to print
  private PrintWriter out;
  
  /***************************************************************************
   * Construct CVCL printer to print on out.
   * @param out the stream on which to print.
   **************************************************************************/
  private CVCLPrinter(PrintWriter out)
  {
    this.out = out;
  }
  
  /***************************************************************************
   * Print tree on out.
   * @param tree the tree to be printed.
   **************************************************************************/
  private void print(AST tree)
  {
    if (tree instanceof Identifier) 
    { 
      printIdentifier((Identifier)tree); 
      return;
    }
    if (tree instanceof Number) 
    { 
      printNumber((Number)tree); 
      return;
    }
    if (tree instanceof Expression) 
    { 
      printExpression((Expression)tree); 
      return;
    }
    if (tree instanceof Type) 
    { 
      printType((Type)tree);
      return;
    }
    if (tree instanceof Declaration) 
    { 
      printDeclaration((Declaration)tree);
      return;
    }
    if (tree instanceof Selector) 
    { 
      printSelector((Selector)tree); 
      return;
    }
    if (tree instanceof TypedIdentifier) 
    { 
      printTypedIdentifier((TypedIdentifier)tree); 
      return;
    }
    if (tree instanceof ValuedIdentifier) 
    { 
      printValuedIdentifier((ValuedIdentifier)tree); 
      return;
    }
    System.out.println("Unknown syntax tree to print: " + tree.toString());
    System.exit(-1);
  }

  /***************************************************************************
   * Print tree on out with parentheses.
   * @param tree the tree to be printed.
   **************************************************************************/
  private void printParens(AST tree)
  {
    out.print('(');
    print(tree);
    out.print(')');
  }
  
  /***************************************************************************
   * Print identifier on out.
   * @param ident the identifier to be printed.
   **************************************************************************/
  private void printIdentifier(Identifier ident)
  {
    if (ident instanceof ValueDeclIdentifier)
    {
      out.print(CVCL.getExternalName((ValueDeclIdentifier)ident));
      return;
    }
    if (ident instanceof Reference)
    {
      Symbol symbol = ((Reference)ident).getSymbol();
      if (symbol != null)
      {
        print(symbol.getIdentifier());
        return;
      }
    }
    out.print(ident.getName());
  }
  
  /***************************************************************************
   * Print number on out.
   * @param number the number to be printed.
   **************************************************************************/
  private void printNumber(Number number)
  {
    out.print(number.getDigits());
  }
  
  /***************************************************************************
   * Print type on out.
   * @param type the type to be printed.
   **************************************************************************/
  private void printType(Type type)
  {
    if (type instanceof AtomicType)
    {
      AtomicType t = (AtomicType)type;
      String name = t.getName();
      if (name.equals("NAT"))
        out.print("INT");
      else
        out.print(name);
      return;
    }
    if (type instanceof FunctionType)
    {
      // CVCL has parsing problems with function types within tuple types
      out.print("(");
      FunctionType t = (FunctionType)type;
      Type[] domain = t.getDomain();
      int n = domain.length;
      Type range = t.getRange();
      if (n == 1)
      {
        if (domain[0] instanceof FunctionType ||
            domain[0] instanceof ArrayType)
          printParens(domain[0]);
        else
          print(domain[0]);
      }
      else
      {
        out.print("(");
        for (int i=0; i<n; i++)
        {
          domain[i].print(out);
          if (i<n-1) out.print(", ");
        }
        out.print(")");
      }
      out.print("->");
      if (range instanceof FunctionType)
        printParens(range);
      else
        print(range);
      out.print(")");
      return;
    }
    if (type instanceof TupleType)
    {
      TupleType t = (TupleType)type;
      Type[] base = t.getBase();
      int n = base.length;
      out.print('[');
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        out.print(", ");
      }
      if (n > 0) print(base[n-1]);
      out.print(']');
      return;
    }
    if (type instanceof RecordType)
    {
      RecordType t = (RecordType)type;
      TypedIdentifier[] base = t.getBase();
      int n = base.length;
      out.print("[#");
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        out.print(", ");
      }
      if (n > 0) print(base[n-1]);
      out.print("#]");
      return;
    }
    if (type instanceof ArrayType)
    {
      ArrayType t = (ArrayType)type;
      Type index = t.getIndex();
      Type base = t.getBase();
      out.print("ARRAY ");
      print(index);
      out.print(" OF ");
      print(base);
      return;
    }
    if (type instanceof SubrangeType)
    {
      // subrange types broken in CVCL
      out.print("INT");
      return;
    }
    if (type instanceof SubType)
    {
      // subtypes broken in CVCL
      SubType t = (SubType)type;
      print(t.getBase());
      return;
    }
    if (type instanceof LetType)
    {
      LetType t = (LetType)type;
      Declaration[] declarations = t.getDeclarations();
      Type base = t.getBase();
      out.print("LET ");
      int n = declarations.length;
      for (int i=0; i<n-1; i++)
      {
        print(declarations[i]);
        out.print(", ");
      }
      if (n > 0) print(declarations[n-1]);
      out.print(" IN ");
      print(base);
      return;
    }
    System.out.println("Unknown type to print: " + type.toString());
    System.exit(-1);
  }
  
  /***************************************************************************
   * Print declaration on out.
   * @param decl the declaration to be printed.
   **************************************************************************/
  private void printDeclaration(Declaration decl)
  {
    // print semantic types, not the syntactic ones,
    // because types have been processed for CVCL.
    Identifier ident = decl.getIdentifier();
    if (decl instanceof TypeDeclaration)
    {
      TypeDeclaration d = (TypeDeclaration)decl;
      Type type = d.getType();
      if (type != null) type = d.getTypeDeclIdentifier().getSymbol().getType();
      print(ident);
      out.print(": TYPE");
      if (type != null)
      {
        out.print(" = ");
        print(type);
      }
      return;
    }
    if (decl instanceof ValueDeclaration)
    {
      ValueDeclaration d = (ValueDeclaration)decl;
      Type type = d.getType();
      if (type != null) 
      {
        ValueSymbol symbol = d.getValueDeclIdentifier().getSymbol();
        TypeExpression te = symbol.getClearType();
        if (te != null)
          type = te.getType();
        else
          type = symbol.getType();
      }
      Expression value = d.getValue();
      print(ident);
      if (type != null)
      {
        out.print(":");
        print(type);
      }
      if (value != null)
      {
        out.print(" = ");
        print(value);
      }
      return;
    }
    // no formula/axiom declarations allowed
    System.out.println("Unknown declaration to print: " + decl.toString());
    System.exit(-1);
  }
  
  /***************************************************************************
   * Print selector on out.
   * @param sel the selector to be printed.
   **************************************************************************/
  private void printSelector(Selector sel)
  {
    if (sel instanceof SelectorIdentifier)
    {
      SelectorIdentifier s = (SelectorIdentifier)sel;
      out.print('.');
      print(s.getIdentifier());
      return;
    }
    if (sel instanceof SelectorIndex)
    {
      SelectorIndex s = (SelectorIndex)sel;
      out.print('[');
      print(s.getIndex());
      out.print(']');
      return;
    }
    if (sel instanceof SelectorNumber)
    {
      SelectorNumber s = (SelectorNumber)sel;
      out.print('.');
      print(s.getNumber());
      return;
    }
    System.out.println("Unknown selector to print: " + sel.toString());
    System.exit(-1);
  }
  
  /***************************************************************************
   * Print typed identifier on out.
   * @param tident the typed identifier to be printed.
   **************************************************************************/
  private void printTypedIdentifier(TypedIdentifier tident)
  {
    print(tident.getIdentifier());
    out.print(":");
    // print semantic type, not the syntactic one, 
    // because type has been preprocessed for CVCL
    print(tident.getIdentifier().getSymbol().getType());
  }
  
  /***************************************************************************
   * Print valued identifier on out.
   * @param vident the valued identifier to be printed.
   **************************************************************************/
  private void printValuedIdentifier(ValuedIdentifier vident)
  {
    print(vident.getIdentifier());
    out.print(":=");
    print(vident.getValue());
  }
  
  /****************************************************************************
   * print expression, use parentheses if priority is greater than treshold
   * @param treshold priority level of enclosing expression
   ***************************************************************************/
  private void printPriority(Expression exp, int treshold)
  {
    int p = exp.getPriority();
    if (p > treshold) out.print("(");
    print(exp);
    if (p > treshold) out.print(")");
  }
  
  /***************************************************************************
   * Print expression on out.
   * @param exp the expression to be printed.
   **************************************************************************/
  private void printExpression(Expression exp)
  {
    if (exp instanceof ApplicationExpression)
    {
      printApplicationExpression((ApplicationExpression)exp);
      return;
    }
    if (exp instanceof Logical)
    {
      Logical e = (Logical)exp;
      if (e.getValue())
        out.print("TRUE");
      else
        out.print("FALSE");
      return;
    }
    if (exp instanceof BinaryExpression)
    {
      printBinaryExpression((BinaryExpression)exp);
      return;
    }
    if (exp instanceof UnaryExpression)
    {
      printUnaryExpression((UnaryExpression)exp);
      return;
    }
    if (exp instanceof IfThenElseExpression)
    {
      IfThenElseExpression e = (IfThenElseExpression)exp;
      out.print("IF ");
      print(e.getCondition());
      out.print(" THEN ");
      print(e.getThenBranch());
      Expression elsebranch = e.getElseBranch();
      if (elsebranch instanceof IfThenElseExpression)
      {
        out.print(" ELS");
        print(elsebranch);
      }
      else
      {
        out.print(" ELSE ");
        print(elsebranch);
      }
      out.print(" ENDIF");
      return;
    }
    if (exp instanceof LetExpression)
    {
      printLetExpression((LetExpression)exp);
      return;
    }
    if (exp instanceof QuantifiedExpression)
    {
      printQuantifiedExpression((QuantifiedExpression)exp);
      return;
    }
    if (exp instanceof RecordTerm)
    {
      RecordTerm e = (RecordTerm)exp;
      ValuedIdentifier[] base = e.getBase();
      int n = base.length;
      out.print("(#");
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        out.print(", ");
      }
      if (n > 0) print(base[n-1]);
      out.print("#)");
      return;
    }
    if (exp instanceof TupleTerm)
    {
      TupleTerm e = (TupleTerm)exp;
      Expression[] base = e.getBase();
      int n = base.length;
      out.print('(');
      for (int i=0; i<n-1; i++)
      {
        print(base[i]);
        out.print(", ");
      }
      if (n > 0) print(base[n-1]);
      out.print(')');
      return;
    }
    if (exp instanceof SelectionTerm)
    {
      SelectionTerm e = (SelectionTerm)exp;
      printPriority(e.getBase(), e.getPriority());
      print(e.getSelector());
      return;
    }
    if (exp instanceof UpdateTerm)
    {
      UpdateTerm e = (UpdateTerm)exp;
      printPriority(e.getBase(), e.getPriority());
      out.print(" WITH ");
      Selector[] selectors = e.getSelectors();
      int n = selectors.length;
      for (int i=0; i<n; i++)
        print(selectors[i]);
      out.print(":=");
      printPriority(e.getUpdate(), e.getPriority());
      return;
    }
    System.out.println("Unknown expression to print: " + exp.toString());
    System.exit(-1);
  }
  
  /****************************************************************************
   * Print application expression.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printApplicationExpression(ApplicationExpression exp)
  {
    Expression fun = exp.getFunction();
    Expression[] args = exp.getArguments();
    printPriority(fun, exp.getPriority());
    int n = args.length; 
    out.print('(');
    for (int i=0; i<n-1; i++)
    {
      print(args[i]);
      out.print(", ");
    }
    if (n > 0) print(args[n-1]);
    out.print(')');
  }
  
  /****************************************************************************
   * Print binary expression.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printBinaryExpression(BinaryExpression exp)
  {
    if (exp instanceof EquivalentFormula)
    {
      // operation is not associative, same predecende as other connectives
      EquivalentFormula e = (EquivalentFormula)exp;
      Expression first = e.getFirst();
      Expression second = e.getSecond();
      if (first instanceof EquivalentFormula ||
          first instanceof NotEquivalentFormula ||
          first instanceof ImpliesFormula)
        printParens(first);
      else
        printPriority(first, e.getPriority());
      out.print(" <=> ");
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula ||
          second instanceof ImpliesFormula)
        printParens(second);
      else
        printPriority(second, e.getPriority());
      return;
    }
    if (exp instanceof ImpliesFormula)
    {
      // operation is right-associative, same precedence as other connectives
      ImpliesFormula e = (ImpliesFormula)exp;
      Expression first = e.getFirst();
      Expression second = e.getSecond();
      if (first instanceof ImpliesFormula ||
          first instanceof EquivalentFormula ||
          first instanceof NotEquivalentFormula)
        printParens(first);
      else
        printPriority(first, e.getPriority());
      out.print(" => ");
      if (second instanceof EquivalentFormula ||
          second instanceof NotEquivalentFormula)
        printParens(second);
      else
        printPriority(second, e.getPriority());
      return;
    }
    if (exp instanceof NotEquivalentFormula)
    {
      // operation is not associative, same precedence as other connectives
      NotEquivalentFormula e = (NotEquivalentFormula)exp;
      Expression first = e.getFirst();
      Expression second = e.getSecond();
      // CVCL 2.0 breaks with XOR
      if (true)
      {
        print(Construct.notFormula(Construct.equivalentFormula(first, second)));
        return;
      }
      if (first instanceof EquivalentFormula || 
          first instanceof NotEquivalentFormula ||
          first instanceof ImpliesFormula)
        printParens(first);
      else
        printPriority(first, e.getPriority());
      out.print(" XOR ");
      if (second instanceof EquivalentFormula || 
          second instanceof NotEquivalentFormula ||
          second instanceof ImpliesFormula)
        printParens(second);
      else
        printPriority(second, e.getPriority());
      return;
    }
    if (exp instanceof PowerTerm)
    {
      // exponentiation is left-associative
      PowerTerm e = (PowerTerm)exp;
      Expression first = e.getFirst();
      Expression second = e.getSecond();
      printPriority(first, e.getPriority());
      out.print("^");
      if (second instanceof PowerTerm)
        printParens(second);
      else
        printPriority(second, e.getPriority());
      return;
    }
    if (exp instanceof DividesTerm)
    {
      DividesTerm e = (DividesTerm)exp;
      Expression first = e.getFirst();
      Expression second = e.getSecond();
      printPriority(first, e.getPriority());
      out.print("/");
      // denominator may need parentheses
      if (second instanceof DividesTerm || second instanceof TimesTerm)
        printParens(second);
      else
        printPriority(second, e.getPriority());
      return;
    }
    String op = null;
    if (exp instanceof AndFormula)
      op = " AND ";
    else if (exp instanceof EqualsFormula)
      op = " = ";
    else if (exp instanceof GreaterEqualFormula)
      op = " >= ";
    else if (exp instanceof GreaterFormula)
      op = " > ";
    else if (exp instanceof LessEqualFormula)
      op = " <= ";
    else if (exp instanceof LessFormula)
      op = " < ";
    else if (exp instanceof MinusTerm)
      op = "-";
    else if (exp instanceof NotEqualsFormula)
      op = " /= ";
    else if (exp instanceof OrFormula)
      op = " OR ";
    else if (exp instanceof PlusTerm)
      op = "+";
    else if (exp instanceof TimesTerm)
      op = "*";
    else
    {
      System.out.println("Unknown binary expression to print: " + exp.toString());
      System.exit(-1);
    }
    BinaryExpression e = (BinaryExpression)exp;
    int p = e.getPriority();
    printPriority(e.getFirst(), p);
    out.print(op);
    printPriority(e.getSecond(), p);
    return;
  }
  
  /***************************************************************************
   * Print unary expression on out.
   * @param exp the expression to be printed.
   **************************************************************************/
  private void printUnaryExpression(UnaryExpression exp)
  {
    if (exp instanceof NegationTerm)
    {
      NegationTerm e = (NegationTerm)exp;
      out.print('-');
      printPriority(e.getBase(), e.getPriority());
      return;
    }
    if (exp instanceof NotFormula)
    {
      NotFormula e = (NotFormula)exp;
      out.print("NOT ");
      printPriority(e.getBase(), e.getPriority());
      return;
    }
    System.out.println("Unknown unary expression to print: " + exp.toString());
    System.exit(-1);
  }
  
  /****************************************************************************
   * Print let expression.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printLetExpression(LetExpression exp)
  {
    ValueDeclaration[] declarations = exp.getDeclarations();
    Expression base = exp.getBase();
    out.print("LET ");
    int n = declarations.length;
    for (int i=0; i<n-1; i++)
    {
      print(declarations[i]);
      out.print(", ");
    }
    if (n > 0) print(declarations[n-1]);
    out.print(" IN ");
    printPriority(base, exp.getPriority());
  }
  
  /****************************************************************************
   * Print quantified expression.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printQuantifiedExpression(QuantifiedExpression exp)
  {
    if (exp instanceof ForallFormula)
    {
      printForallFormula((ForallFormula)exp);
      return;
    }
    if (exp instanceof ExistsFormula)
    {
      printExistsFormula((ExistsFormula)exp);
      return;
    }
    printHead(exp);
    printPriority(exp.getBase(), exp.getPriority());
  }
  
  /****************************************************************************
   * Print forall formula.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printForallFormula(ForallFormula exp)
  {
    CVCLFormulaMask mask = State.getFormulaMask();
    CVCLPrinter printer = this;
    if (mask != null) printer = new CVCLPrinter(new PrintWriter(new StringWriter()));
    Expression constraint = printer.printHead(exp);
    Expression base = exp.getBase();
    if (constraint != null) base = Construct.impliesFormula(constraint, base); 
    printer.printPriority(base, exp.getPriority());
    if (printer != this) print(mask.getMask(exp));
  }
  
  /****************************************************************************
   * Print exists formula.
   * @param exp the expression to be printed.
   ***************************************************************************/
  private void printExistsFormula(ExistsFormula exp)
  {
    CVCLFormulaMask mask = State.getFormulaMask();
    CVCLPrinter printer = this;
    if (mask != null) printer = new CVCLPrinter(new PrintWriter(new StringWriter()));
    Expression constraint = printer.printHead(exp);
    Expression base = exp.getBase();
    if (constraint != null) base = Construct.andFormula(constraint, base); 
    printer.printPriority(base, exp.getPriority());
    if (printer != this) print(mask.getMask(exp));
  }
    
  /***************************************************************************
   * Prints quantifier/variables part of expression.
   * Returns collected constraint (null, if none).
   * @param out the stream on which the text is written
   * @return the constaint.
   **************************************************************************/
  private Expression printHead(QuantifiedExpression exp)
  {
    if (exp instanceof ForallFormula)
      out.print("FORALL");
    else if (exp instanceof ExistsFormula)
      out.print("EXISTS");
    else if (exp instanceof ArrayTerm)
      out.print("ARRAY");
    else if (exp instanceof LambdaTerm)
      out.print("LAMBDA");
    else
    {
      System.out.println("Unknown quantified expression to print: " + 
        exp.toString());
      System.exit(-1);
    }
    TypedIdentifier[] variables = exp.getVariables();
    out.print('(');
    int n = variables.length;
    Expression constraint = null;
    for (int i=0; i<n; i++)
    {
      ValueDeclIdentifier ident = variables[i].getIdentifier();
      AST type0 = variables[i].getType();
      AST type1 = null;
      if (i+1 < n) type1 = variables[i+1].getType();
      if (type0 == type1)
        print(ident);
      else  
        print(variables[i]);
      TypeExpression te = ident.getSymbol().getClearType();
      if (te != null && te.getExpression() != null)
      {
        if (constraint == null)
          constraint = te.getExpression();
        else
          constraint = Construct.andFormula(constraint, te.getExpression());
      }
      if (i < n-1)
        out.print(", ");
      else
        out.print("): "); 
    }
    return constraint;
  }
}
