// ----------------------------------------------------------------------------
// ProverReasoning.java
// Reasoning with the help of an external prover.
// $Id: ProverReasoning.java,v 1.27 2007/10/23 20:10:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Proving;

import fmrisc.ProofNavigator.*;
import fmrisc.Semantics.*;
import fmrisc.Syntax.*;

import java.util.*;

/**
 * 
 * Reasoning with the help of an external prover.
 *
 */
public final class ProverReasoning
{
  // remember whether current prover tries hard to simplify
  // (set when prover is created)
  private static boolean hard;
  
  // remember whether current prover generates counterexamples
  // (set on demand only)
  private static boolean generate = false;
  
  /***************************************************************************
   * Simplify a single formula in some proof state.
   * @param state a proof state.
   * @param formula a formula of this proof state to be simplified.
   * @param hard true iff simplification is attempted hard.
   * @return true iff everything went okay.
   **************************************************************************/
  public static boolean simplify(ProofState state, Formula formula, 
    boolean hard)
  {
    // signal not to generate counterexamples
    ProverReasoning.generate = false;
    
    // environment for type checking formulas read from prover
    // (also prover initialization depends on it)
    Environment env = state.getEnvironment();
    Environment oenv = State.getEnvironment();
    State.setEnvironment(env);
    
    // create prover for simplification
    Prover prover = createProver(state, hard);
    if (prover == null) 
    {
      State.setEnvironment(oenv);
      return false;
    }
    
    // simplify formula hard in empty context
    State.setRunningProver(prover);
    simplify(prover, formula, true);
    State.setRunningProver(null);
    
    // skip the other simplifications
    //prover.terminate();
    //if (true) 
    //{
    //  State.setEnvironment(oenv);
    //  return true;
    //}
    
    // assert assumptions
    Formula[] as = state.getAssumptions();
    int an = as.length;
    for (int i=0; i<an; i++)
    {
      Formula f = as[i];
      if (f != formula) 
        prover.assertFormula(f.getExpression());
    }
    
    // assert negated goals
    Formula[] gs = state.getGoals();
    int gn = gs.length;
    for (int i=0; i<gn; i++)
    {
      Formula f = gs[i];
      if (f != formula) 
        prover.assertFormula(Construct.notFormula(f.getExpression()));
    }
    
    // simplify formula in asserted context
    simplify(prover, formula, hard);
    prover.terminate();

    // we are done
    State.setEnvironment(oenv);
    return true;
  }
  
  /***************************************************************************
   * Simplify the current proof state and potentially close it.
   * @param hard true if hard (and time-consuming) simplification is tried.
   * @param generate true if counterexample is to be generated,
   *         if state could not be closed.
   * @return true iff everything went okay.
   *************************************************************************/
  public static boolean simplify(ProofState state, boolean hard,
    boolean generate)
  { 
    // signal to generate counterexamples
    ProverReasoning.generate = generate;
    
    // construct formula sequence 
    Formula[] as = state.getAssumptions();
    Formula[] gs = state.getGoals();
    int an = as.length;
    int gn = gs.length;
    int n = an+gn;
    Formula[] formulas = new Formula[n];
    for (int i=0; i<an; i++)
    {
      Formula f = as[i];
      formulas[i] = f;
    }
    for (int i=0; i<gn; i++)
    {
      Formula f = gs[i];
      formulas[an+i] = f;
    }
    
    // environment for type checking formulas read from prover
    // (also prover initialization depends on it)
    Environment env = state.getEnvironment();
    Environment oenv = State.getEnvironment();
    State.setEnvironment(env);
    
    // create prover for simplification
    Prover prover = createProver(state, hard);
    if (prover == null) 
    {
      State.setEnvironment(oenv);
      return false;
    }
    State.setRunningProver(prover);
    
    // simplify formulas in empty context
    for (int i=0; i<n; i++)
      simplify(prover, formulas[i], hard);
    
    // state may be closed or simplified
    Answer answer = isInconsistent(prover, formulas, an);
    if (answer instanceof ValidAnswer)
      state.close(answer.getJustification());
    else if (!State.isAborted())
    {
      if (answer instanceof InvalidAnswer)
      {
        InvalidAnswer ianswer = (InvalidAnswer)answer;
        Expression[] counterExample = ianswer.getCounterExample();
        if (counterExample != null)
        {
          int m = counterExample.length;
          for (int i=0; i<m; i++)
          {
            if (Checking.process(counterExample[i]) == null)
            {
              // reset environment and terminate prover
              State.setRunningProver(null);
              State.setEnvironment(oenv);
              prover.terminate();
              return false;
            }
          }
          state.setCounterExample(counterExample);
        }
      }
      // simplify proof state
      simplify(prover, formulas, an, 0, n-1);
    }
      
    // reset environment and terminate prover
    State.setRunningProver(null);
    State.setEnvironment(oenv);
    prover.terminate();
    return true;
  }
  
  /****************************************************************************
   * Create prover to deal with denoted state.
   * @param state a proof state.
   * @param hard true iff prover shall use hard simplification (no masking)
   * @return a prover initialized with the declarations and axioms of 
   *         the state (null, if an error occured)
   **************************************************************************/
  private static Prover createProver(ProofState state, boolean hard)
  {
    // remember hardness state
    ProverReasoning.hard = hard;
    
    // prover does not timeout and masks quantified formulas
    Prover prover = State.newProver(State.getProverLog(), 0, !hard);
    if (!prover.isOkay())
    {
      State.setError("error in starting prover: " + prover.getError());
      return null;
    }

    // add declarations of atomic types to prover state
    Environment env = state.getEnvironment();
    Iterator symbols = env.getTypeSymbols().iterator();
    while (symbols.hasNext())
    {
      TypeSymbol symbol = (TypeSymbol)symbols.next();
      Type type = symbol.getType();
      if (!(type instanceof AtomicType)) continue;
      TypeDeclIdentifier ident = symbol.getTypeDeclIdentifier();
      AtomicType atype = (AtomicType)type;
      if (!ident.getName().equals(atype.getName())) continue;
      TypeDeclaration decl = Construct.typeDeclaration(ident, null);
      prover.addDeclaration(decl);
    }
    
    // add declarations of values to prover state
    // must be in the order in which they were entered
    // since a value may have a subtype depending on another value
    Vector svector = env.getGlobalValueSymbols();
    int sn = svector.size();
    for (int i=0; i<sn; i++)
    {
      ValueSymbol symbol = (ValueSymbol)svector.elementAt(i);
      ValueDeclaration decl = 
        Construct.valueDeclaration(symbol.getValueDeclIdentifier(),
                                   symbol.getType(), null);
      prover.addDeclaration(decl);
    }
          
    // simplify axioms and add them to prover state 
    // only soft simplification because type-checking axioms may be nasty
    Formula[] axioms = state.getAxioms();
    int axn = axioms.length;
    for (int i=0; i<axn; i++)
    {
      Formula f = axioms[i];
      simplify(prover, f, false);
      prover.assertAxiom(f.getExpression());
    }
    
    // prover is initialized
    return prover;
  }
  
  /****************************************************************************
   * Simplify formula in current prover context.
   * @param prover the prover to be used.
   * @param formula the formula to be simplified.
   * @param hard true if prover shall try hard simplification.
   ***************************************************************************/
  private static void simplify(Prover prover, Formula formula, boolean hard)
  {
    if (State.isAborted()) return;
    prover.tryHard(hard);
    prover.pushContext();
    Answer answer = prover.queryFormula(formula.getExpression());
    prover.popContext();
    Expression result;
    if (answer instanceof ValidAnswer)
      result = Construct.logical(true);
    else if (State.useAutoSimplify())
    { 
      prover.pushContext();
      result = prover.simplifyExpression(formula.getExpression());
      prover.popContext();
      if (result == null) 
      {
        State.getSystemOut().println("Error in simplification of " + 
          formula.getExpression().toString());
        if (State.getError() != null) 
        {
          State.getSystemOut().println(State.getError());
          State.resetError();
        }
        return;
      }
    }
    else
      result = formula.getExpression();
    prover.tryHard(ProverReasoning.hard);
    formula.setExpression(result);
  }
 
  /***************************************************************************
   * Assert to prover context to formulas
   * @param prover the prover to be used.
   * @param formulas the formulas of the context consisting of anumber
   *        assumptions followed by a sequence of goals.
   * @param anumber the number of assumptions in formulas.
   * @param from the start of the context range (inclusive).
   * @parem end the end of the context range (inclusive).
   * @return the answer.
   *************************************************************************/
  private static void assertContext(Prover prover,
    Formula[] formulas, int anumber, int from, int to)
  {
    for (int i=from; i<=to; i++)
    {
      if (i < anumber)
        prover.assertFormula(formulas[i].getExpression());
      else
        prover.assertFormula(Construct.notFormula(formulas[i].getExpression()));
    }
  }
  
  /***************************************************************************
   * Simplify every formula in formulas[from..to] under the assumption that
   * every formula in context outside the range of [from..to] is true.
   * The prover operates in a context where all formulas outside the range 
   * context[from..to] are already assumed as true.
   * @param prover the prover to use in the simplification
   * @param formulas the formulas to be simplified
   * @param context the formulas representing the context (same length)
   * @param from the lower bound of the simplification range
   * @param to the upper bound of the simplification range
   *************************************************************************/
  private static void simplify(Prover prover, 
    Formula[] formulas, int anumber, int from, int to)
  {
    if (from > to) return;
    if (from == to) 
    {      
      simplify(prover, formulas[from], hard);
      return;
    }
    int mid = (from+to)/2;
    
    prover.pushContext();
    assertContext(prover, formulas, anumber, mid+1, to);
    simplify(prover, formulas, anumber, from, mid);
    prover.popContext();
    
    prover.pushContext();
    assertContext(prover, formulas, anumber, from, mid);
    simplify(prover, formulas, anumber, mid+1, to);
    prover.popContext();
  }
  
  /***************************************************************************
   * Check whether context is inconsistent.
   * @param prover the prover to be used.
   * @param formulas the formulas of the context consisting of a sequence
   *        of assumptions followed by a sequence of goals.
   * @param anumber the number of assumptions in formulas .
   * @return the answer.
   *************************************************************************/
  private static Answer isInconsistent(Prover prover, 
    Formula[] formulas, int anumber)
  {
    // during proof replay, not needed to check for inconsistency
    // of state with children
    ProofState[] replayChildren = State.getReplayChildren();
    if (replayChildren != null && replayChildren.length > 0) 
      return new InvalidAnswer("not attempted");
    prover.pushContext();
    assertContext(prover, formulas, anumber, 0, formulas.length-1);
    prover.generateCounterExamples(generate);
    Answer answer = prover.queryFormula(Construct.logical(false));
    prover.generateCounterExamples(false);
    prover.popContext();
    return answer;
  }
}
