// ----------------------------------------------------------------------------
// Context.java
// Handling of formula contexts.
// $Id: Context.java,v 1.12 2006/06/23 21:30:01 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Semantics;

import fmrisc.Syntax.*;
import fmrisc.ProofNavigator.*;
import java.util.*;

/**
 * 
 * handling of formula contexts
 *
 */
public final class Context
{
  // the representation of a nested context as a stack of formula sequences
  private Stack cstack = new Stack(); // context stack
  private Vector ctable = null;       // current top of context stack
  
  // the formulas collected in this context as a stack of formula sequences
  private Stack fstack = new Stack(); // formula stack
  private Expression ftop = null;     // current top of formula stack
  
  /***************************************************************************
   * construct a new context empty of assumptions
   **************************************************************************/
  public Context()
  {
    open();
  }
  
  /**************************************************************************
   * return the formula currently on the top of the stack
   * @return the formula on the top of the stack
   *************************************************************************/
  public Expression getFormula()
  {
    if (ftop == null) return null;
    return ftop.instantiate();
  }
  
  /**************************************************************************
   * add an assumption to the current (sub)context
   * @param formula the assumption to be added
   *************************************************************************/
  public void assume(Expression formula)
  {
    ctable.add(formula);
  }
  
  /***************************************************************************
   * conjoin formula to the current formula on the top of the stack
   * @param formula formula to be conjoined to top of stack
   **************************************************************************/
  public void implies(Expression formula)
  {
    if (ftop == null) 
      ftop = formula;
    else
      ftop = Construct.andFormula(ftop, formula);
    fstack.pop();
    fstack.push(ftop);
  }
  
  /****************************************************************************
   * open a new subcontext
   * subcontexts may be arbitrarily nested; every subcontext must be
   * ultimately either aborted or commited.
   * if the subcontext is aborted, every formula created in it 
   * by implies() is discarded.
   * if the subcontext is commited, every formula created in it
   * by implies() is added to the parent context.
   **************************************************************************/
  public void open()
  {
    ctable = new Vector();
    cstack.push(ctable);
    ftop = null;
    fstack.push(ftop);
  }
  
  /***************************************************************************
   * return to the parent context discarding all the formulas that were added
   * to the subcontext since the last call of open().
   * all assumptions added since the last call of open() are discarded.
   **************************************************************************/
  public void abort()
  {
    cstack.pop();
    ctable = (Vector)cstack.peek();
    fstack.pop();
    ftop = (Expression)fstack.peek();
  }
  
  /***************************************************************************
   * create implication of (conjunction of) context and top of formula stack.
   * @return the implication (or NULL, if th top of formula stack is NULL)
   **************************************************************************/
  private Expression combine()
  {
    if (ftop == null) return null;
    
    // construct hypothesis
    Expression hypothesis = null;
    int n = ctable.size();
    for (int i=0; i<n; i++)
    { 
      Expression f = (Expression)ctable.elementAt(i);
      if (hypothesis == null)
        hypothesis = f;
      else
        hypothesis = Construct.andFormula(hypothesis, f);
    }
        
    // construct implication
    Expression formula = ftop;
    if (hypothesis != null)
      formula = Construct.impliesFormula(hypothesis, formula);
    return formula;
  }
  
  /***************************************************************************
   * return to the parent context and add to it all the formulas that were
   * added to the subcontext since the last call of open().
   * all assumptions added since the last call of open() are discarded.
   **************************************************************************/
  public void commit()
  {
    Expression formula = combine();
    
    cstack.pop();
    ctable = (Vector)cstack.peek();
    
    fstack.pop();
    ftop = (Expression)fstack.peek();
    if (formula == null) return;
    if (ftop == null)
      ftop = formula;
    else
      ftop = Construct.andFormula(ftop, formula);
    fstack.pop();
    fstack.push(formula);
  }
  
  /***************************************************************************
   * close formula on top of formula stack with respect to current scope.
   **************************************************************************/
  public void close()
  {
    if (ftop == null) return;
    
    // close formula
    Collection free = FreeVariables.compute(ftop);
    Vector tids = new Vector();
    Collection symbols = State.getEnvironment().getValueSymbols();
    Iterator iterator = symbols.iterator();
    while (iterator.hasNext())
    {
      ValueSymbol symbol = (ValueSymbol)iterator.next();
      if (!free.contains(symbol)) continue;
      tids.addElement(new TypedIdentifier(symbol.getValueDeclIdentifier(), 
                                          symbol.getType()));
    }
    if (tids.size() == 0) return;

    // add closed formula, still needs to be instantiated
    ftop = Construct.forallFormula(
      ASTUtil.toTypedIdentifierArray(tids), ftop);
    fstack.pop();
    fstack.push(ftop);
  }
}
