// ----------------------------------------------------------------------------
// IfThenElseExpression.java
// Handling of if-then-else expressions.
// $Id: IfThenElseExpression.java,v 1.13 2006/03/01 17:12:57 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import java.io.*;

/**
 * Handling of if-then-else expressions.
 */
public final class IfThenElseExpression extends ExpressionBase
{
  private Expression condition;  // the selection condition
  private Expression thenbranch; // the then branch
  private Expression elsebranch; // the else branch

  /****************************************************************************
   * construct if-then-else expressions from condition, thenbranch, elsebranch.
   * @param condition the formula deciding which branch is selected
   * @param thenbranch the branch selected if the formula yields true
   ***************************************************************************/
  public IfThenElseExpression(Expression condition, 
    Expression thenbranch, Expression elsebranch)
  {
    this.condition = condition;
    this.thenbranch = thenbranch;
    this.elsebranch = elsebranch;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 0;
  }
  
  /***************************************************************************
   * get condition of expression
   * @return the condition
   **************************************************************************/
  public Expression getCondition()
  {
    return condition;
  }
  
  /***************************************************************************
   * get then branch of expression
   * @return the then branch
   **************************************************************************/
  public Expression getThenBranch()
  {
    return thenbranch;
  }
  
  /***************************************************************************
   * get else branch of expression
   * @return the else branch
   **************************************************************************/
  public Expression getElseBranch()
  {
    return elsebranch;
  }
  
  /***************************************************************************
   * Prints text representation of tree on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print("IF ");
    condition.print(out);
    out.print(" THEN ");
    thenbranch.print(out);
    if (elsebranch instanceof IfThenElseExpression)
    {
      out.print(" ELS");
      ((IfThenElseExpression)elsebranch).printCore(out);
    }
    else
    {
      out.print(" ELSE ");
      elsebranch.print(out);
    }
    out.print(" ENDIF");
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    return Construct.ifThenElseExpression(
      getCondition().instantiate(),
      getThenBranch().instantiate(),
      getElseBranch().instantiate());
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST condition0 = condition.accept(visitor);
    AST thenbranch0 = thenbranch.accept(visitor);
    AST elsebranch0 = elsebranch.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.ifThenElseExpression(
      (Expression)condition0, (Expression)thenbranch0, (Expression)elsebranch0);
  }
}
