// ----------------------------------------------------------------------------
// TupleTerm.java
// Handling of tuple terms.
// $Id: TupleTerm.java,v 1.11 2006/02/23 19:51:06 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import java.io.*;

/**
 * Handling of tuple terms.
 */
public final class TupleTerm extends ExpressionBase
{
  // the tuple components
  private Expression[] base;
 
  /****************************************************************************
  * construct tuple term of base terms
  * @param base the array of base terms, base.length > 1
   ***************************************************************************/
  public TupleTerm(Expression[] base)
  {
    this.base = base;
  }
  
  /***************************************************************************
   * get binding level for printing (lower numbers mean greater binding power)
   * @return the priority level
   **************************************************************************/
  public int getPriority()
  {
    return 0;
  }
  
  /***************************************************************************
   * get base expressions
   * @return the base expression sequence
   **************************************************************************/
  public Expression[] getBase()
  {
    return base;
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    int n = base.length;
    out.print('(');
    for (int i=0; i<n-1; i++)
    {
      base[i].print(out);
      out.print(", ");
    }
    base[n-1].print(out);
    out.print(')');
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * Overrides ASTBase.print such that parentheses are never printed
   * around the tuple.
   * @param out the stream on which the text is written
   **************************************************************************/
  public void print(PrintWriter out)
  {
    printCore(out);
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Expression instantiate()
  {
    Expression[] base = getBase();
    int n = base.length;
    Expression[] base0 = new Expression[n];
    for (int i=0; i<n; i++)
      base0[i] = base[i].instantiate();
    return Construct.tupleTerm(base0);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null).
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    int n = base.length;
    Expression[] base0 = new Expression[n];
    for (int i=0; i<n; i++)
      base0[i]= (Expression)base[i].accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.tupleTerm(base0);
  }
}
