// ----------------------------------------------------------------------------
// TypeDeclaration.java
// Handling of type declarations.
// $Id: TypeDeclaration.java,v 1.12 2006/02/23 19:51:04 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.Syntax;

import java.io.*;

import fmrisc.Semantics.Symbol;

/**
 * Handling of type declarations.
 */
public final class TypeDeclaration extends DeclarationBase
{
  private TypeDeclIdentifier name; // the declared name
  private Type type; // the type associated to the name
  
  /****************************************************************************
   * construct type declaration with denoted name and type.
   * @param name the declared name
   * @param type the type associated to the name (may be null)
   ***************************************************************************/
  public TypeDeclaration(TypeDeclIdentifier name, Type type)
  {
    this.name = name;
    this.type = type;
  }
  
  /****************************************************************************
   * returns declaration name
   * @return the name
   ***************************************************************************/
  public TypeDeclIdentifier getTypeDeclIdentifier()
  {
    return name;
  }
  
  /****************************************************************************
   * returns declaration name
   * @return the name
   ***************************************************************************/
  public Identifier getIdentifier()
  {
    return name;
  }
  
  /****************************************************************************
   * returns declaration type
   * @return the type (may be null)
   ***************************************************************************/
  public Type getType()
  {
    return type;
  }
  
  /***************************************************************************
   * create instantiated copy of declaration
   * @return the instantiated copy
   **************************************************************************/
  public TypeDeclaration instantiate()
  {
    TypeDeclIdentifier ident = getTypeDeclIdentifier();
    TypeDeclIdentifier ident0 = ident.instantiate();
    Type type = getType();
    Type type0 = null;
    if (type != null) type0 = type.instantiateType();
    return Construct.typeDeclaration(ident0, type0);
  }
  
  /***************************************************************************
   * Prints text representation of tree on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    name.print(out);
    out.print(": TYPE");
    if (type != null)
    {
      out.print(" = ");
      type.print(out);
    }
  }
  
  /****************************************************************************
   * returns symbol
   * @return the symbol (null, if not yet determined)
   ***************************************************************************/
  public Symbol getSymbol()
  {
    return name.getSymbol();
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null)
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST name0 = name.accept(visitor);
    AST type0 = null;
    if (type != null) type0 = type.accept(visitor);
    if (!visitor.isCloned()) return null;
    return Construct.typeDeclaration((TypeDeclIdentifier)name0, (Type)type0);
  }
}
